# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

"""Splits raw sequencing FASTQs by barcode, quality filters, and annotated UMIs.

Usage:
    dupliganger barcode-split-quality-filter-umi-anno [options] <barcode_file> <in.fq> <in.barcode.fq> <in.umi.fq>
    dupliganger barcode-split-quality-filter-umi-anno [options] <barcode_file> <in.R1.fq> <in.R2.fq> <in.barcode.fq> <in.umi.fq>

Input:
    This utility works on FASTQ files produced by Illumina's bcl2fastq2.  It
    assumes that it was a paired-end sequencing run with the following files:
        * <in.R1.fq> - Read 1.
        * <in.R2.fq> - Read 2.
        * <in.barcode.fq> - The barcode for the read pair.
        * <in.umi.fq> - The UMI of the barcode.
    You'll need to speak with your sequencing facility to determine which files
    are which.  For example, our facility has produced files that look like this:
        * Undetermined_S0_R1_* - This is Read 1.
        * Undetermined_S0_R2_* - This is the barcode.
        * Undetermined_S0_R3_* - This is the UMI.
        * Undetermined_S0_R4_* - This is Read 2.
    You'll also need to provide a list of barcodes used (<barcode_file>).  See
    below.

Barcodes and Output File Names:
    You'll need to pass a list of expected barcodes in <barcode_file>.  Put one
    barcode per line in this file; alternatively, you may put a barcode
    followed by whitespace followed by a sample name (no spaces in sample names
    allowed). If you take the former approach, output files will be named after
    their barcodes; if you take the latter approach, output files will be named
    after your samples.  Reads that don't match any barcode are placed in a
    rejects file.

UMI quality and quailty filtering:
    If an UMI has an 'N' in its sequence, this utility will reject it and write
    it to the rejects files.

    Additionally, by default, Dupligänger quality filters out any read that
    has an average quality score less than 30 across a window size of 1. You
    can change those parameters with -w and -q respectively.  Alternatively,
    you can disable UMI quality filtering altogether by setting either -w or -q
    to '0'.

Note:
    Dupligänger supports (and autodetects) input FASTQ files that are gzipped.

Options:
    -h, --help
    -v, --verbose               Be verbose.
    -o OUT_DIR                  Place results in directory OUT_DIR.
    --phred PHRED               How FASTQ phred scores are encoded [default: 33].
                                Default is '33'. Use '64' to specify phred64
                                encoding.
    -w W, --umi-qf-win-size W   Set window size of 'W' nucleotides for UMI
                                quality filtering. See notes above. Disable
                                quality filtering by setting -w or -q to 0.
                                [default: 1].
    -q Q, --umi-min-qual Q      Set minimum average quality score 'Q' for UMI
                                quality filtering across window size 'W'
                                nucleotides for UMI.  See notes above. Disable
                                quality filtering by setting -w or -q to 0.
                                [default: 30].
    -T, --no-delete-tmp-files   Upon some failure, do *not* delete the
                                temporary files created by this utility.
    -r, --no-write-rejects      Do not write files of reads rejected due to
                                mismatching barcodes, 'N's in UMIs, or
                                low-quality UMIs.
"""

###############
### Imports ###
###############

# Python 3 imports
from __future__ import absolute_import
from __future__ import division
from builtins import range

# Dupligänger imports
from dupliganger.constants import *
from dupliganger.exceptions import *
from dupliganger.common import (args_to_out_dir, pgopen, is_gzipped, gzwrite,
        tmpf_start, tmpf_finish, tmpf_open, filename_in_to_out_fqgz)

## Other imports
from docopt import docopt

# for opening up files (e.g. open_out_pe_barcode_fastq_files)
import contextlib

# For filename fixing
import os

# For writing gzipped files
import gzip

# For checking presence of illumina index under python 2/3.
import io

# For shell-like "which()"
try:
    from shutil import which
except ImportError:
    from whichcraft import which

# For function partials
import functools


#################
### Constants ###
#################


#################
### Functions ###
#################

def parse_barcode_list_file(barcode_list_file):
    """Parse the barcodes list file. It can come in several different formats.
    Autodetect which format, and return a dict mapping barcode to optional
    sample name.  If sample names are not given, all vals are ...????.

    Example format 1:
        ACGTGGGG
        TTAATTCC

    Example format 2:
        ACGTGGGG    SampleName1
        TTAATTCC    SampleName2

    Example format 3:
        SampleName1     ACGTGGGG
        SampleName2     TTAATTCC

    Args:
        barcode_list_file (str): Name of the barcodes list file.
    Returns:
        (dict): keys: barcodes (str), vals: None (None) or sample name (str).
    """
    lines = []
    with pgopen(1, barcode_list_file) as bcf:
        for line in bcf:
            line = line.strip()
            if line == '' or line[0] == '#':
                # skip blank lines and comments
                continue
            lines.append(line)

    # Determine number of columns in file (and confirm all lines have same
    # number of columns)
    num_cols = set()
    for line in lines:
        num_cols.add(len(line.strip().split()))
    if len(num_cols) == 0:
        raise CannotContinueException(
                """ERR808: Barcodes list file is empty?""")
    elif len(num_cols) > 1:
        raise CannotContinueException(
                """ERR801: Cannot parse barcodes file, mix of number of columns?""")
    num_cols = list(num_cols)[0]

    # keys: barcodes (str), vals: None (None) or sample name (str).
    barcodes_list_dict = {}
    if num_cols == 1:
        # Example format 1
        for bc in lines:
            barcodes_list_dict[bc] = None
    elif num_cols == 2:
        # Example format 2 or 3.  Which one?
        chars1 = ''
        chars2 = ''
        barcodes_in_col = None
        # First, figure out which column barcodes are in.
        for line in lines:
            c1, c2 = line.split()
            chars1 += c1
            chars2 += c2
        chars1 = chars1.strip('ACGT')
        chars2 = chars2.strip('ACGT')
        if (chars1 == '' and chars2 == ''):
            raise CannotContinueException(
                    "ERR678: Cannot parse barcodes file. Did you put barcodes "
                    "in both columns?  If using two columns, place barcodes in "
                    "one column, and sample names in the other column.")
        elif chars1 != '' and chars2 != '':
            raise CannotContinueException(
                    "ERR876: Cannot parse barcodes file. Please put barcodes in "
                    "one column and sample names in another column. Please note "
                    "that no spaces are allowed in sample names.")
        elif chars1 == '' and chars2 != '':
            barcodes_in_col = 1
        elif chars1 != '' and chars2 == '':
            barcodes_in_col = 2

        # Now parse it
        if barcodes_in_col == 1:
            # Example format 2
            for line in lines:
                bc, name = line.split()
                barcodes_list_dict[bc] = name
        else:
            # Example format 3
            for line in lines:
                name, bc = line.split()
                barcodes_list_dict[bc] = name

    return barcodes_list_dict

@contextlib.contextmanager
def open_out_pe_barcode_fastq_files(outdir, delete_temp_files_upon_failure,
        fp_write, barcode_list_dict):
    """Context manager to help manage the variable number of output paired-end
    FASTQ files.  Takes as input a list of barcodes.  Based upon the list of
    barcodes (or optionally mapped sample names if provided in
    barcode_list_dict), does the following:
        * opens *temporary* paired-end output files for writing, two for each
          barcode.
        * yields those back for writing
        * when finished writing, closes the files.
        * renames the temporary files to be named after barcodes (or sample
          names if provided).

    Args:
        outdir (str): Place output fastq files in directory outdir.
        delete_temp_files_upon_failure (bool): If there is a failure and this
            arg is True, delete the temp files.
        fp_write (function): Function pointer for writing output files.
        barcode_list_dict (dict): Keys are expected barcodes, vals are either
            None or sample names.
    Yields:
        (dict): Each key is a barcode, each value is a two-tuple of file handles
            for that barcode (first fh is R1 output file for that barcode,
            second fh is for R2).
    """
    # keys: barcodes, vals: [read1_tmp_filehandle, read2_tmp_filehandle]
    output_files = {}
    tmp_output_filenames = {}

    try:
        ## This 'try' block: open output files for writing, yield.

        for barcode, opt_sample_name in barcode_list_dict.items():
            # First, determine if we're using barcodes or sample names for
            # final file names ('pretend' is here just so we can reuse
            # filename_in_to_out_fqgz...)
            if opt_sample_name is not None:
                in_pretend_r1 = "{}.R1.fq".format(opt_sample_name)
                in_pretend_r2 = "{}.R2.fq".format(opt_sample_name)
            else:
                in_pretend_r1 = "{}.R1.fq".format(barcode)
                in_pretend_r2 = "{}.R2.fq".format(barcode)

            # Come up with final filenames, open temp versions of those files,
            # record file names for later use....
            out_r1 = filename_in_to_out_fqgz(in_pretend_r1, SUFFIX_REMOVE_UMI,
                    False, outdir)
            out_r2 = filename_in_to_out_fqgz(in_pretend_r2, SUFFIX_REMOVE_UMI,
                    False, outdir)
            tmp_out_r1, tmp_out_r2 = tmpf_start(out_r1, out_r2)

            tmp_output_filenames[barcode] = (tmp_out_r1, tmp_out_r2)
            output_files[barcode] = (fp_write(tmp_out_r1), fp_write(tmp_out_r2))

        yield output_files

    except Exception as e:
        ## This 'except' block: try to recover from failure (close/delete files)
        for barcode in barcode_list_dict:
            # Close tmp files first
            for out_f in output_files[barcode]:
                try:
                    out_f.close()
                except Exception:
                    pass
            if delete_temp_files_upon_failure:
                for tmp_filename in tmp_output_filenames[barcode]:
                    try:
                        os.remove(tmp_filename)
                    except Exception:
                        pass
        raise e

    finally:
        ## This 'finally' block: normal exit, close/rename files.
        for barcode in barcode_list_dict:
            # Close tmp files first
            for out_f in output_files[barcode]:
                out_f.close()
            # Rename from tmp to final...
            tmp_out_r1, tmp_out_r2 = tmp_output_filenames[barcode]
            try:
                tmpf_finish(tmp_out_r1, tmp_out_r2)
            except Exception:
                # We catch and ignore this exception so as to not mask previous exceptions.
                pass

def split_qf_umi_anno_raw_pe(in_r1, in_r2, in_bc, in_umi, out_rejects_r1,
        out_rejects_r2, out_rejects_bc, out_rejects_umi, out_files,
        barcodes_list_dict, phred, min_umi_qual, umi_qf_win_size,
        write_rejects_files):
    """Using non-barcode-split (ie 'raw') FASTQ files from Illumina's
    bcl2fastq2 (paired-end version) as input, split by barcode, then quality
    filter on UMI, then create paired FASTQ output files with read name
    annotated with UMI.

    ASSUMPTION:
        All reads ID lines have two fields and look something like this:
          @EAS139:136:FC706VJ:2:5:1000:12850 1:N:0

    Args:
        in_r1 (file): Read1 input fastq file handle.
        in_r2 (file): Read2 input fastq file handle.
        in_bc (file): FASTQ file handle of barcodes.
        in_umi (file): FASTQ file handle of UMIs.
        out_rejects_r1 (file): Read1 rejects output fastq file handle.
        out_rejects_r2 (file): Read2 rejects output fastq file handle.
        out_rejects_bc (file): FASTQ output file handle of barcodes (rejects).
        out_rejects_umi (file): FASTQ output file handle of UMIs (rejects).
        out_files ([file]): Output paired-end fastq file handles.
        barcodes_list_dict (dict): Keys: barcodes, vals: optional sample name.
        phred (int): How quality scores are phred encoded (33 or 64).
        write_rejects_files (bool): Whether or not to write rejects files.
    """
    fq_rec = "{}\n{}\n+\n{}\n"

    # For report
    hist_barcodes = {}
    hist_umis = {}
    count_barcode_mismatch = 0
    count_umi_has_N = 0
    count_umi_failed_qf = 0

    # Assume that you've already setup a bunch of stuff already...
    while True:
        name1, seq1, _, qual1 = \
                (in_r1.readline().rstrip() for i in range(4))
        name2, seq2, _, qual2 = \
                (in_r2.readline().rstrip() for i in range(4))
        name_bc, seq_bc, _, qual_bc = \
                (in_bc.readline().rstrip() for i in range(4))
        name_umi, seq_umi, _, qual_umi = \
                (in_umi.readline().rstrip() for i in range(4))

        if not name1:
            # EOF
            break

        # record for report
        if seq_bc not in barcodes_list_dict:
            count_barcode_mismatch += 1
        elif 'N' in seq_umi:
            count_umi_has_N += 1

        if seq_bc not in barcodes_list_dict or 'N' in seq_umi:
            # Barcode doesn't match, or 'N' in UMI sequence. Write to rejects
            # files.

            if write_rejects_files:
                out_rejects_r1.write(fq_rec.format(name1, seq1, qual1))
                out_rejects_r2.write(fq_rec.format(name2, seq2, qual2))
                out_rejects_bc.write(fq_rec.format(name_bc, seq_bc, qual_bc))
                out_rejects_umi.write(fq_rec.format(name_umi, seq_umi, qual_umi))
            continue
        else:
            # Begin quality filter UMI.
            umi_failed_qf = False
            if min_umi_qual > 0 and umi_qf_win_size > 0:
                start = 0
                stop = umi_qf_win_size
                len_qual_umi = len(qual_umi)
                while stop <= len_qual_umi and not umi_failed_qf:
                    qual_sum = 0
                    for char in qual_umi[start:stop]:
                        qual_sum += ord(char) - phred
                    if (qual_sum/(stop-start) < min_umi_qual):
                        umi_failed_qf = True
                        break
                    start += 1
                    stop += 1
            # End quality filter

            if umi_failed_qf:
                # UMI failed quality filter. Write to rejects files.
                count_umi_failed_qf += 1
                if write_rejects_files:
                    out_rejects_r1.write(fq_rec.format(name1, seq1, qual1))
                    out_rejects_r2.write(fq_rec.format(name2, seq2, qual2))
                    out_rejects_bc.write(fq_rec.format(name_bc, seq_bc, qual_bc))
                    out_rejects_umi.write(fq_rec.format(name_umi, seq_umi, qual_umi))
                continue

        # Read pair has good barcode and UMI. Continue along...

        # record for report
        hist_barcodes.setdefault(seq_bc, 0)
        hist_barcodes[seq_bc] += 1
        hist_umis.setdefault(seq_umi, 0)
        hist_umis[seq_umi] += 1

        # Split the ID line into just the name and the 'info' field.
        # ASSUMPTION: All reads ID lines have two fields and look something
        # like this:
        #       @EAS139:136:FC706VJ:2:5:1000:12850 1:N:0
        just_name1, info1 = name1.split()
        just_name2, info2 = name2.split()
        just_name_bc, _ = name_bc.split()
        just_name_umi, _ = name_umi.split()

        # Confirm names are same.
        if just_name1 != just_name2:
            raise ParseException(
                    'R1 read name != R2 read name! ({} != {})'.format(
                        just_name1, just_name2))
        if just_name1 != just_name_bc:
            raise ParseException(
                    'R1 read name != barcode read name! ({} != {})'.format(
                        just_name1, just_name_bc))
        if just_name1 != just_name_umi:
            raise ParseException(
                    'R1 read name != UMI read name! ({} != {})'.format(
                        just_name1, just_name_umi))

        # Build the records
        name_anno = "{}{}{}".format(just_name1, DELIM_ANNO, seq_umi)
        info_anno1 = "{}:{}".format(info1, seq_bc)
        info_anno2 = "{}:{}".format(info2, seq_bc)
        record1 = "{} {}\n{}\n+\n{}\n".format(name_anno, info_anno1, seq1,
                qual1)
        record2 = "{} {}\n{}\n+\n{}\n".format(name_anno, info_anno2, seq2,
                qual2)

        out_files[seq_bc][0].write(record1)
        out_files[seq_bc][1].write(record2)

    return (hist_barcodes, hist_umis, count_barcode_mismatch, count_umi_has_N,
            count_umi_failed_qf)

def parse_args(args):
    """Parse the command line arguments. """

    umi_reads_file = os.path.expanduser(args['<in.umi.fq>'])
    barcode_reads_file = os.path.expanduser(args['<in.barcode.fq>'])
    barcode_list_file = os.path.expanduser(args['<barcode_file>'])
    delete_temp_files_upon_failure = not args['--no-delete-tmp-files']
    write_rejects_files = not args['--no-write-rejects']
    min_umi_qual = int(args['--umi-min-qual'])
    umi_qf_win_size = int(args['--umi-qf-win-size'])


    # phred
    if args['--phred'] not in ('33', '64'):
        raise ArgumentException("ERR212: --phred can be only '33' or '64'.")
    phred = int(args['--phred'])

    # Paired or single end?
    if args['<in.R2.fq>']:
        # mode = PE
        paired = True
        # Convert ~ to real path
        args['<in.R1.fq>'] = os.path.expanduser(args['<in.R1.fq>'])
        args['<in.R2.fq>'] = os.path.expanduser(args['<in.R2.fq>'])
        reads_files = [args['<in.R1.fq>'], args['<in.R2.fq>']]
    else:
        # Convert ~ to real path
        args['<in.fq>'] = os.path.expanduser(args['<in.fq>'])

        # mode = SR
        paired = False
        reads_files = [args['<in.fq>']]

    fp_write = functools.partial(open, mode = 'w')

    # Return an appropriate function pointer for annotation.
    if paired:
        fp_split_qf_umi_anno_raw = split_qf_umi_anno_raw_pe
    else:
        fp_split_qf_umi_anno_raw = split_qf_umi_anno_raw_sr

    outdir = args_to_out_dir(args)

    return (fp_split_qf_umi_anno_raw, fp_write, outdir, phred, min_umi_qual,
            umi_qf_win_size, write_rejects_files,
            delete_temp_files_upon_failure, barcode_list_file,
            barcode_reads_file, umi_reads_file, reads_files)

def run(fp_split_qf_umi_anno_raw, fp_write, outdir, phred, min_umi_qual,
        umi_qf_win_size, write_rejects_files, delete_temp_files_upon_failure,
        barcode_list_file, barcode_reads_file, umi_reads_file, reads_files):
    """Start the run.

    Args:
        fp_split_qs_umi_anno (Function): Function to be used to parse the FASTQ
            file(s) and produce output.
        fp_write (Function): Function to be used to write output file(s).
        outdir (str): Output directory for results
        phred (int): How quality scores are phred encoded (33 or 64).
        force_paired (bool): Whether user wants to force paired-end, even if
            if we didn't detect it.
        barcode_list_file (str): File of list of barcodes.
        barcode_reads_file (str): FASTQ file of barcodes.
        umi_reads_file (str): FASTQ file of UMIs.
        reads_files ([str]): Array of read or read pair FASTQ files to be parsed.
    """

    # Parse barcode_list_file, build barcode_list_dict.
    barcodes_list_dict = parse_barcode_list_file(barcode_list_file)

    # Some datasets have look like this:
    #       @NS500451:139:H5TV5AFXX:1:11101:3928:1111 1:N:0:ATCACGTT
    # Others like this:
    #       @NS500451:139:H5TV5AFXX:1:11101:3928:1111
    # Detect.
    if is_gzipped(reads_files[0]):
        with gzip.open(reads_files[0], mode='rt') as in1:
            first_line = in1.readline()
    else:
        with io.open(reads_files[0], mode='r', encoding='latin-1') as in1:
            first_line = in1.readline()
    has_index = True if len(first_line.split()) > 1 else False

    reports = None
    common_file_prefix = None
    out_files = []

    if len(reads_files) == 2:
        # paired-end
        infile1, infile2 = reads_files

        common_file_prefix = os.path.commonprefix((infile1, infile2))

        # Build up rejects filenames
        out_rejects_r1 = filename_in_to_out_fqgz(infile1, SUFFIX_REJECTS,
                False, outdir)
        out_rejects_r2 = filename_in_to_out_fqgz(infile2, SUFFIX_REJECTS,
                False, outdir)
        out_rejects_bc = filename_in_to_out_fqgz(barcode_reads_file,
                SUFFIX_REJECTS, False, outdir)
        out_rejects_umi = filename_in_to_out_fqgz(umi_reads_file,
                SUFFIX_REJECTS, False, outdir)

        out_files += [out_rejects_r1, out_rejects_r2, out_rejects_bc,
                out_rejects_umi]

        # quick alias
        rm = delete_temp_files_upon_failure
        # open files for reading and writing...
        with    pgopen(1, infile1) as in_r1, \
                pgopen(1, infile2) as in_r2, \
                pgopen(1, barcode_reads_file) as in_bc, \
                pgopen(1, umi_reads_file) as in_umi, \
                tmpf_open(fp_write, rm, out_rejects_r1) as tmp_out_rejects_r1, \
                tmpf_open(fp_write, rm, out_rejects_r2) as tmp_out_rejects_r2, \
                tmpf_open(fp_write, rm, out_rejects_bc) as tmp_out_rejects_bc, \
                tmpf_open(fp_write, rm, out_rejects_umi) as tmp_out_rejects_umi, \
                open_out_pe_barcode_fastq_files(outdir,
                        delete_temp_files_upon_failure, fp_write,
                        barcodes_list_dict) as barcode_out_files:

            reports = fp_split_qf_umi_anno_raw(in_r1, in_r2, in_bc, in_umi,
                    tmp_out_rejects_r1, tmp_out_rejects_r2, tmp_out_rejects_bc,
                    tmp_out_rejects_umi, barcode_out_files, barcodes_list_dict, phred,
                    min_umi_qual, umi_qf_win_size, write_rejects_files)

            (hist_barcodes, hist_umis, count_barcode_mismatch, count_umi_has_N,
                    count_umi_failed_qf) = reports

            # Translate/Record actual barcode out_files for tests
            for fhs in barcode_out_files.values():
                # a little hacky here...
                for fh in fhs:
                    tmp_filename = fh.name
                    size = len(".tmp.") + TMP_FILE_NAME_RANDOM_STR_SIZE
                    filename = tmp_filename[:-size]
                    out_files.append(filename)

    elif len(reads_files) == 1:
        # single-end
        raise Exception('NOT YET IMPLEMENTED!!! ')
    else:
        raise ControlFlowException(
                """ERR911: Not possible to be here.""")

    # write report file
    report_file = "{}.report".format(common_file_prefix)
    with open(report_file, 'w') as f:
        f.write("num_reads_with_barcode_mismatch: {}\n".format(
            count_barcode_mismatch))
        f.write("num_reads_with_Ns_in_UMI: {}\n".format(
            count_umi_has_N))
        f.write("num_reads_with_umi_poor_quality: {}\n".format(
            count_umi_failed_qf))
        f.write("Barcode_Histogram:\n")
        for bc in sorted(hist_barcodes.keys()):
            f.write("{}: {}\n".format(bc, hist_barcodes[bc]))
        f.write("UMI_Histogram:\n")
        for umi in sorted(hist_umis.keys()):
            f.write("{}: {}\n".format(umi, hist_umis[umi]))

    return out_files

###############
### Classes ###
###############


############
### Main ###
############

def main():
    args = docopt(__doc__)
    run(*parse_args(args))

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

# Python 3 imports
from __future__ import absolute_import
from __future__ import division

# NOTE: Do *not* do the following:
# from builtins import str
# from builtins import chr
# py-lmdb uses bytes() for py3 and str() for py2.
# Doing any of the above starts turning things into unicode.  We want native
# strings.

################
### Defaults ###
################

# The configuration file
DEFAULT_DUPLIGANGER_CONFIG_FILE =                    "dupliganger.config"

#################
### Constants ###
#################

# Bioo kit name
KIT_BIOO =                                                           'bioo'

# Bioo UMIs
UMIS_BIOO = ['AACGCCAT', 'AAGGTACG', 'AATTCCGG', 'ACACAGAG', 'ACACTCAG',
        'ACACTGTG', 'ACAGGACA', 'ACCTGTAG', 'ACGAAGGT', 'ACGACTTG', 'ACGTCAAC',
        'ACGTCATG', 'ACTGTCAG', 'ACTGTGAC', 'AGACACTC', 'AGAGGAGA', 'AGCATCGT',
        'AGCATGGA', 'AGCTACCA', 'AGCTCTAG', 'AGGACAAC', 'AGGACATG', 'AGGTTGCT',
        'AGTCGAGA', 'AGTGCTGT', 'ATAAGCGG', 'ATCCATGG', 'ATCGAACC', 'ATCGCGTA',
        'ATCGTTGG', 'CAACGATC', 'CAACGTTG', 'CAACTGGT', 'CAAGTCGT', 'CACACACA',
        'CAGTACTG', 'CATCAGCA', 'CATCGTTC', 'CCAAGGTT', 'CCTAGCTT', 'CGATTACG',
        'CGCCTATT', 'CGTTCCAT', 'CGTTGGAT', 'CTACGTTC', 'CTACTCGT', 'CTAGAGGA',
        'CTAGGAAG', 'CTAGGTAC', 'CTCAGTCT', 'CTGACTGA', 'CTGAGTGT', 'CTGATGTG',
        'CTGTTCAC', 'CTTCGTTG', 'GAACAGGT', 'GAAGACCA', 'GAAGTGCA', 'GACATGAG',
        'GAGAAGAG', 'GAGAAGTC', 'GATCCTAG', 'GATGTCGT', 'GCCGATAT', 'GCCGATTA',
        'GCGGTATT', 'GGAATTGG', 'GGATAACG', 'GGCCTAAT', 'GGCGTATT', 'GTCTTGTC',
        'GTGATGAG', 'GTGATGTC', 'GTGTACTG', 'GTGTAGTC', 'GTTCACCT', 'GTTCTGCT',
        'GTTGTCGA', 'TACGAACC', 'TAGCAAGG', 'TAGCTAGC', 'TAGGTTCG', 'TATAGCGC',
        'TCAGGACT', 'TCCACATC', 'TCGACTTC', 'TCGTAGGT', 'TCGTCATC', 'TGAGACTC',
        'TGAGAGTG', 'TGAGTGAG', 'TGCTTGGA', 'TGGAGTAG', 'TGTGTGTG', 'TTCGCCTA',
        'TTCGTTCG']

# Dupligänger defined SAM TAGs
SAM_TAG_READ1_HAMMING_DIST_UMI =                                    'd1:i:'
SAM_TAG_READ2_HAMMING_DIST_UMI =                                    'd2:i:'
SAM_TAG_READ1_NUM_CLOSEST_UMIS =                                    'n1:i:'
SAM_TAG_READ2_NUM_CLOSEST_UMIS =                                    'n2:i:'
SAM_TAG_READ1_CLOSEST_UMI      =                                    'c1:i:'
SAM_TAG_READ2_CLOSEST_UMI      =                                    'c2:i:'

# ReportDB metrics

LOG_NUM_LOCATIONS =                                         'num_locations'
LOG_NUM_UNIQUE_UMI_AND_LOCATION_COMBINATIONS = (
                                 'num_unique_umi_and_location_combinations')
LOG_NUM_READ_GROUPS =                                     'num_read_groups'
LOG_NUM_DUP_GROUPS =                                       'num_dup_groups'
LOG_NUM_READ_GROUPS_WITH_UMI_ERROR =       'num_read_groups_with_umi_error'
LOG_NUM_READ_GROUPS_WITH_UMI_ERROR_DIST = (
                                      'num_read_groups_with_umi_error_dist')
LOG_NUM_READ_GROUPS_REJECTED_DUE_TO_UMI_ERROR_DIST = (
                           'num_read_groups_rejected_due_to_umi_error_dist')
# ReportDB
REPORT_DB_COUNT_METRICS = [
    LOG_NUM_LOCATIONS,
    LOG_NUM_UNIQUE_UMI_AND_LOCATION_COMBINATIONS,
    LOG_NUM_READ_GROUPS,
    LOG_NUM_DUP_GROUPS,
    LOG_NUM_READ_GROUPS_WITH_UMI_ERROR, ]
REPORT_DB_8_UMI_DIST_COUNT_METRICS = [
    LOG_NUM_READ_GROUPS_WITH_UMI_ERROR_DIST,
    LOG_NUM_READ_GROUPS_REJECTED_DUE_TO_UMI_ERROR_DIST, ]


# Pad the read_group_id with zeros until you have this many digits.
READ_GROUP_ID_DIGITS =                                                   10
## Delimiters for the read name annotations ##
# Delimiter to separate read names from their annotations
DELIM_ANNO = '-'
# Delimiter to separate read pairs
DELIM_ANNO_READ_PAIR = '^'
# Delimiter to separate records of different types
DELIM_ANNO_TYPE = ';'
# Delimiter to seperate records of the same type
DELIM_ANNO_ELEMENT = ','

## Other delimiters
# Delimiter to separate SAM file fields
DELIM_SAM_FIELD = '\t'

# Delimiter to separate elements of a list in a SimpleBucketLmdb
# Note: chr(30) is ACSII "record separator"
DELIM_SAM_LIST_LMDB = chr(30)

# Default delimiter to separate items in a SimpleBucketDb
DELIM_BUCKET_LIST = ','

# Delimiter to separate a ReadGroup's metadata from the SAM records the
# ReadGroup contains.
# Note: chr(29) is ACSII "group separator"
DELIM_READ_GROUP_METADATA_SAM_RECORDS = chr(29)

## Database stuff
STORE_OPTION_LMDB = 'lmdb'
STORE_OPTION_MEMORY = 'memory'

### LMDB stuff ###
# 1TB database size
LMDB_DB_SIZE = 1024**4
LMDB_MAX_DBS = 10
RECORDS_PER_TXN = 1000000
RECORDS_PER_TXN = 100000
# RECORDS_PER_TXN = 1
# RECORDS_PER_TXN = 5000000
# RECORDS_PER_TXN = 31
# RECORDS_PER_TXN = 1
# RECORDS_PER_TXN = 3
# RECORDS_PER_TXN = 500

# File suffixes
# For FASTQ reads that are rejected (e.g. UMI quality score too low)
SUFFIX_REJECTS =                                                  'rejects'
# After --remove-umi is run.
SUFFIX_REMOVE_UMI =                                                 'rmumi'
# After remove-adapter is run.
SUFFIX_REMOVE_ADAPTER =                                           'rmadapt'
# For the "too-short" files produced by remove-adapter.
SUFFIX_REMOVE_ADAPTER_TOO_SHORT =                                'tooshort'
# After --qtrim is run.
SUFFIX_QTRIM =                                                      'qtrim'
SUFFIX_QTRIM_UNPAIRED =                                          'unpaired'
# After --annotate-qtrim umi is run.
SUFFIX_ANNOTATE_QTRIM =                                              'anno'

# A 'PCR or optical duplicate' is marked with 0x400 in the FLAG field of SAM
# files.
SAM_FORMAT_FLAG_DUPLICATE = 0x400

# Number of characters in the random string of
TMP_FILE_NAME_RANDOM_STR_SIZE =                                           6

# Default random seed
RANDOM_SEED =                                                'Little Ashes'


# Dupligänger log filename
LOG_FILENAME =                                           'dupliganger.log'
# Overall log level
LOG_LEVEL =                                                         'DEBUG'
# Log level for dupliganger.log
LOG_LEVEL_FOR_FILE =                                                 'INFO'
# Logging configuration dict
LOGGING = {
    'version': 1,
    'handlers': {
        #'console': {
        #    'class': 'logging.StreamHandler',
        #    'level': LOG_LEVEL,
        #},
        'file': {
            'class': 'logging.FileHandler',
            'level': LOG_LEVEL_FOR_FILE,
            'filename': LOG_FILENAME,
            'mode': 'w',
        }
    },
    'root': {
        'level': LOG_LEVEL,
        #'handlers': ['console', 'file']
        'handlers': ['file']
    },
}

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

###############
### Imports ###
###############

# Python 2/3 compatibility imports
from __future__ import absolute_import, division, print_function

# NOTE: Do *not* do the following:
# from builtins import str, chr, object
# py-lmdb uses bytes() for py3 and str() for py2.
# This package has different code for py2 and py3.
# And importing that future 'object' has a bug that screws up __slots__ in
# py2 (causes different behavior than in py3).

# For iterating over dicts in py23
from future.utils import iteritems

# Dupligänger imports
from dupliganger.constants import *
from dupliganger.exceptions import *

### DEBUGGING
from dupliganger.sam import (to_location_key_with_5p_trimming)

## Other imports

# For filename fixing
import os

# For writing errors about hard clipping
import sys

# For function partials
import functools

# For annotation database
import lmdb

# # For database stuff
import contextlib

#################
### Constants ###
#################


#################
### Functions ###
#################


###############
### Classes ###
###############

class ParentDb(object):
    pass

class ParentDbLmdb(ParentDb):
    """Parent database object.  Basically just wraps py-lmdb's Environment object."""
    __slots__ = ('env', )
    def __init__(self, filename, max_dbs, map_size):
        self.env = lmdb.open(filename, max_dbs=max_dbs, map_size=map_size)
    @contextlib.contextmanager
    def begin(self, write=False):
        with self.env.begin(write=write) as txn:
            yield txn

class ParentDbDict(ParentDb):
    """Let our python dict databases act like our LMDB databases."""
    @contextlib.contextmanager
    def begin(self, junk=False):
        yield TxnDict()

class TxnDict(object):
    """Dummy class.  Does nothing except help others not complain."""

class SimpleObjectDb(object):
    """A simple database in which a string key points to a string value which
    is converted to an object by the class passed to this database upon
    initialization.
    """
    __slots__ = ('db', 'item_class')

    def __str__(self):
        """String representation of this db. For testing. Produces a
        dictionary-like output, sorted by key."""
        s = ''
        for k, v in sorted(iteritems(self.to_dict())):
            s += '{}: {}\n'.format(k,v)
        return s

class SimpleObjectDbLmdb(SimpleObjectDb):

    def __init__(self, db_name, parent_db, env, item_class):
        """
         Note: The item_class needs to implement the load() and the __repr__()
             method, which loads the bytes from the LMDB database.  In other
             words, the item_class defines and is entirely responsible for its
             string/bytes representation in the LMDB, and conversion back and
             forth.
         Args:
            db_name (str): Name of the database.
            parent_db (ParentDbLmdb): The LMDB parent database.
            env (lmdb.Environment): An LMDB environment.
            item_class (type): The class of item being stored.
        """
        if sys.version_info.major == 3:
            self.db = env.open_db(bytes(db_name, encoding='latin-1'))
        elif sys.version_info.major == 2:
            self.db = env.open_db(db_name)
        self.parent_db = parent_db
        self.item_class = item_class

    def to_dict(self):
        """Build a dictionary version of this LMDB. For testing."""
        d = {}
        with self.parent_db.begin() as txn:
                for key, val in self.iteritems(txn):
                    d[key] = val
        return d

    def fp_get(self, txn):
        return functools.partial(self.get, txn)
    def fp_put(self, txn):
        return functools.partial(self.put, txn)

    if sys.version_info.major == 3:

        def iteritems(self, txn):
            """Returns an items iterator."""
            for key, val in iter(txn.cursor(self.db)):
                yield (key.decode('latin-1'), self.convert(val))

        def get(self, txn, obj_id):
            """Cousin of convert()"""
            item_bytes = txn.get(bytes(obj_id, encoding='latin-1'), db=self.db)
            return self.item_class().load(item_bytes.decode())

        def put(self, txn, obj_id, item):
            """Python3 version of put.
            Args:
                obj_id (str): The ID of the item.
            """
            obj_id_bytes = bytes((obj_id), encoding='latin-1')
            item_bytes = bytes(str(item), encoding='latin-1')
            txn.put(obj_id_bytes, item_bytes, db=self.db)

        def convert(self, item_bytes):
            """Cousin of get(). Python3 version of convert()."""
            return self.item_class().load(item_bytes.decode())

    elif sys.version_info.major == 2:

        def iteritems(self, txn):
            """Returns an items iterator."""
            for key, val in iter(txn.cursor(self.db)):
                yield (key, self.convert(val))

        def get(self, txn, obj_id):
            """Cousin of convert()"""
            item_bytes = txn.get(obj_id, db=self.db)
            return self.item_class().load(item_bytes)

        def put(self, txn, obj_id, item):
            """Python2 version of put.
            Args:
                obj_id (str): The ID of the item.
            """
            item_str = repr(item)
            txn.put(obj_id, item_str, db=self.db)

        def convert(self, item_repr):
            """Cousin of get(). Python2 version of convert()."""
            return self.item_class().load(item_repr)

class SimpleObjectDbDict(SimpleObjectDb):
    def __init__(self, _1=None, _2=None, _3=None):
        self.db = {}
    def iteritems(self, _):
        """Returns an items iterator. (Note: the extra iter() call is necessary
        to make this work in py3.)"""
        return iter(iteritems(self.db))
    def fp_get(self, _):
        return functools.partial(self.get, _)
    def fp_put(self, _):
        return functools.partial(self.put, _)
    def get(self, _, obj_id):
        return self.db.get(obj_id)
    def put(self, _, obj_id, item):
        self.db[obj_id] = item
    def convert(self, thing):
        return thing
    def to_dict(self):
        return self.db

class SimpleBucket(object):
    """A simple database, with string keys (string cheese?), which point at a
    simple list of members, separated by a delimiter (e.g. ',' or '\t')."""
    __slots__ = ('db', 'delim', 'item_class', 'db_name')
    def __str__(self):
        """String representation of this db. For testing. Produces a
        dictionary-like output, sorted by key."""
        s = ''
        for k, v in sorted(iteritems(self.to_dict())):
            s += '{}: {}\n'.format(k,v)
        return s

class SimpleBucketLmdb(SimpleBucket):
    def __init__(self, db_name, parent_db, env, delim, item_class):
        """
         Args:
            db_name (str): Name of the database. (For debugging purposes only I
                believe.)
            parent_db (ParentDbLmdb): The LMDB parent database.
            env (lmdb.Environment): An LMDB environment.
            delim (str): The delimiter to separate values.
            item_class (type): The class of item being stored.
        """
        if sys.version_info.major == 3:
            self.db = env.open_db(bytes(db_name, encoding='latin-1'))
        elif sys.version_info.major == 2:
            self.db = env.open_db(db_name)
        self.parent_db = parent_db
        self.delim = delim
        self.item_class = item_class
        # Next line just for occassional debuggin.
        self.db_name = db_name

    def to_dict(self):
        """Build a dictionary version of this LMDB. For testing."""
        d = {}
        with self.parent_db.begin() as txn:
                for key, val in self.iteritems(txn):
                    d[key] = val
        return d

    def fp_get(self, txn):
        return functools.partial(self.get, txn)
    def fp_put(self, txn):
        return functools.partial(self.put, txn)
    def fp_append(self, txn):
        return functools.partial(self.append, txn)
    def fp_append_many(self, txn):
        return functools.partial(self.append_many, txn)

    if sys.version_info.major == 3:

        def iteritems(self, txn):
            """Returns an items iterator."""
            for key, val in iter(txn.cursor(self.db)):
                yield (key.decode('latin-1'), self.convert(val))

        def get(self, txn, bucket_id):
            items_str = txn.get(bytes(bucket_id, encoding='latin-1'),
                    db=self.db)
            if items_str is None:
                return None
            elif self.item_class == str:
                return items_str.decode().split(self.delim)
            else:
                return [ self.item_class(item) for item in
                        items_str.decode().split(self.delim) ]

        def put(self, txn, bucket_id, items):
            assert type(items) is list or type(items) is tuple
            bucket_id_bytes = bytes(bucket_id, encoding='latin-1')
            items_bytes = bytes(self.delim.join([str(item) for item in items]),
                    encoding='latin-1')
            txn.put(bucket_id_bytes, items_bytes, db=self.db)

        def convert(self, items_concat):
            """
            Args:
                items_concat (bytes): Items as bytes, concatenated together.
            """
            items_str = items_concat.decode().split(self.delim)
            if self.item_class == str:
                return items_str
            else:
                return [ self.item_class(item) for item in items_str ]

        def append(self, txn, bucket_id, item):
            bucket_id_bytes = bytes(bucket_id, encoding='latin-1')
            existing = txn.get(bucket_id_bytes, db=self.db)
            if existing:
                items_concat = bytes(self.delim.join((existing.decode(),
                    str(item))), encoding='latin-1')
            else:
                items_concat = bytes(item, encoding='latin-1')
            txn.put(bucket_id_bytes, items_concat, db=self.db)

        def append_many(self, txn, bucket_id, items):
            bucket_id_bytes = bytes(bucket_id, encoding='latin-1')
            existing = txn.get(bucket_id_bytes, db=self.db)
            if existing:
                items_concat = bytes(self.delim.join((existing.decode(),
                    self.delim.join(items))), encoding='latin-1')
            else:
                items_concat = bytes(self.delim.join(items),
                        encoding='latin-1')
            txn.put(bucket_id_bytes, bytes(items_concat), db=self.db)

    elif sys.version_info.major == 2:

        def iteritems(self, txn):
            """Returns an items iterator."""
            for key, val in iter(txn.cursor(self.db)):
                yield (key, self.convert(val))

        def get(self, txn, bucket_id):
            items_str = txn.get(bucket_id, db=self.db)
            if items_str is None:
                return None
            elif self.item_class == str:
                return items_str.split(self.delim)
            else:
                return [ self.item_class(item) for item in
                        items_str.split(self.delim) ]

        def put(self, txn, bucket_id, items):
            assert type(items) is list or type(items) is tuple
            items_repr = self.delim.join([str(item) for item in items])
            txn.put(bucket_id, items_repr, db=self.db)

        def convert(self, items_concat):
            """
            Args:
                items_concat (str): Items as str repr, concatenated together.
            """
            items_str = items_concat.split(self.delim)
            if self.item_class == str:
                return items_str
            else:
                return [ self.item_class(item) for item in items_str ]

        def append(self, txn, bucket_id, item):
            """Py2 version of append()"""
            existing = txn.get(bucket_id, db=self.db)
            if existing:
                items_concat = self.delim.join((existing, str(item)))
            else:
                items_concat = item
            txn.put(bucket_id, items_concat, db=self.db)

        def append_many(self, txn, bucket_id, items):
            existing = txn.get(bucket_id, db=self.db)
            if existing:
                items_concat = self.delim.join((existing, self.delim.join(items)))
            else:
                items_concat = self.delim.join(items)
            txn.put(bucket_id, items_concat, db=self.db)

class SimpleBucketDict(SimpleBucket):
    def __init__(self, _1=None, _2=None, _3=None):
        self.db = {}
    def iteritems(self, _):
        """Returns an items iterator. (Note: the extra iter() call is necessary
        to make this work in python3.)"""
        return iter(iteritems(self.db))
    def to_dict(self):
        return self.db
    def fp_get(self, _):
        return functools.partial(self.get, _)
    def fp_put(self, _):
        return functools.partial(self.put, _)
    def fp_append(self, _):
        return functools.partial(self.append, _)
    def fp_append_many(self, _):
        return functools.partial(self.append_many, _)
    def get(self, _, bucket_id):
        return self.db.get(bucket_id)
    def put(self, _, bucket_id, items):
        assert type(items) is list or type(items) is tuple
        self.db[bucket_id] = items
    def append(self, _, bucket_id, item):
        self.db.setdefault(bucket_id, [])
        self.db[bucket_id].append(item)
    def append_many(self, _, bucket_id, items):
        self.db.setdefault(bucket_id, [])
        self.db[bucket_id] += items

    def convert(self, thing):
        return thing

class LocationBucketDb(object):
    """A location database, mostly just a simple wrapper around a) a
    SimpleBucket, and b) a function pointer that specifies how to turn a
    ReadGroup into a location_key to be used for lookups in this db.
    """

    __slots__ = 'db fp_to_location_key'.split()

    def __init__(self, db, fp_to_location_key):
        """
         Args:
            db (SimpleBucket): Store for this database wrapper..
            fp_to_location_key (function): A function that converts a read
                group (i.e. list of Read objects) to a genomic location to be
                used as the key to the LocationBucketDb.
        """
        self.db = db
        self.fp_to_location_key = fp_to_location_key

    def __str__(self):
        """String representation of the underlying db. For testing. Produces a
        dictionary-like output, sorted by key."""
        return str(self.db)

    def iteritems(self, txn):
        """Returns an items iterator."""
        return self.db.iteritems(txn)

    def fp_append(self, txn):
        return functools.partial(self.append, txn)

    def append(self, txn, read_group_id, read_group):
        try:
            loc_key = self.fp_to_location_key(read_group)
            self.db.append(txn, loc_key, read_group_id)
        except HardClippingNotSupportedException:
            m = ("WARNING: Encountered hard-clipped read(s). Hard clipping "
                " not supported. Dropping read(s). Here's the read(s): {}\n"
                .format(read_group))
            sys.stderr.write(m)

class DebugMultipleLocationBucketDbs(object):
    """Just like LocationBucketDb, except it's many in one. FOR DEBUGGING."""

    __slots__ = 'db1 db2'.split()

    def __init__(self, simple_bucket_class, env, delim, item_class):
        self.db1 = simple_bucket_class('debug_location_bucket_store1', env, delim,
                item_class)
        self.db2 = simple_bucket_class('debug_location_bucket_store2', env, delim,
                item_class)

    def fp_append(self, txn):
        return functools.partial(self.append, txn)

    def append(self, txn, read_group_id, read_group):
        loc_key1 = to_location_key_with_5p_trimming(read_group)
        self.db1.append(txn, loc_key1, read_group_id)


############
### Main ###
############

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from os import path
from codecs import open  # To use a consistent encoding
import sys

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# List of required packages for python 2 and 3.
install_requires=['docopt', 'lmdb', 'distance', 'future', 'psutil']

# List of backported required packages for python2.
if sys.version_info[0] == 2:
    install_requires += [
            # For lru_cache
            'functools32',
            # For unix-like "which" tool.
            'whichcraft']

setup(
    name='dupliganger',
    version='0.93',

    description="A reference-based, UMI-aware, 5ʹ-trimming-aware PCR duplicate removal pipeline.",
    long_description=long_description,

    # Dupligänger's homepage.
    url='https://github.com/uoregon-postlethwait/dupliganger',

    # Author details
    author='Postlethwait Lab, University of Oregon',
    author_email='postlethwait.lab@gmail.com',

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        # Who Dupligänger is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',

        # License
        'License :: Free for non-commercial use',

        # Python versions supported.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    keywords='bioinformatics',
    zip_safe=True,

    # Packages provided.
    packages=['dupliganger'],

    # Required packages.
    install_requires=install_requires,

    # Require Python 2.7 or any versions of Python 3 starting with 3.4:
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',

    # Reference the License
    license = 'LICENSE.txt',

    # The dupliganger command line tool.
    entry_points={
        'console_scripts': [
            'dupliganger=dupliganger.command_line:main_wrapper',
        ],
    },

    # URLs
    project_urls={
        'Source': 'https://github.com/uoregon-postlethwait/dupliganger',
        'Bug Reports': 'https://github.com/uoregon-postlethwait/dupliganger/issues',
    },
)

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

import pytest
from pytest import fixture
from dupliganger.sam import parse_cigar
from dupliganger.common import HardClippingNotSupportedException

class TestParseCigar(object):
    """Test parse_cigar()."""

    #############
    ### Tests ###
    #############

    def test_simple(self):
        assert parse_cigar(100, '+', '1M') == (100, 100, 100, 100)
        assert parse_cigar(100, '-', '1M') == (100, 100, 100, 100)
        assert parse_cigar(100, '+', '101M') == (100, 100, 200, 200)
        assert parse_cigar(100, '-', '101M') == (200, 200, 100, 100)

    def test_left_soft_clip(self):
        assert parse_cigar(100, '+', '1S1M') == (99, 100, 100, 100)
        assert parse_cigar(100, '-', '1S1M') == (100, 100, 100, 99)
        assert parse_cigar(100, '+', '1S101M') == (99, 100, 200, 200)
        assert parse_cigar(100, '-', '1S101M') == (200, 200, 100, 99)

        assert parse_cigar(100, '+', '10S1M') == (90, 100, 100, 100)
        assert parse_cigar(100, '-', '10S1M') == (100, 100, 100, 90)
        assert parse_cigar(100, '+', '10S101M') == (90, 100, 200, 200)
        assert parse_cigar(100, '-', '10S101M') == (200, 200, 100, 90)

    def test_right_soft_clip(self):
        assert parse_cigar(100, '+', '1M1S') == (100, 100, 100, 101)
        assert parse_cigar(100, '-', '1M1S') == (101, 100, 100, 100)
        assert parse_cigar(100, '+', '101M1S') == (100, 100, 200, 201)
        assert parse_cigar(100, '-', '101M1S') == (201, 200, 100, 100)

        assert parse_cigar(100, '+', '1M10S') == (100, 100, 100, 110)
        assert parse_cigar(100, '-', '1M10S') == (110, 100, 100, 100)
        assert parse_cigar(100, '+', '101M10S') == (100, 100, 200, 210)
        assert parse_cigar(100, '-', '101M10S') == (210, 200, 100, 100)

    def test_left_and_right_soft_clip(self):
        assert parse_cigar(100, '+', '1S1M1S') == (99, 100, 100, 101)
        assert parse_cigar(100, '-', '1S1M1S') == (101, 100, 100, 99)
        assert parse_cigar(100, '+', '1S101M1S') == (99, 100, 200, 201)
        assert parse_cigar(100, '-', '1S101M1S') == (201, 200, 100, 99)

        assert parse_cigar(100, '+', '10S1M10S') == (90, 100, 100, 110)
        assert parse_cigar(100, '-', '10S1M10S') == (110, 100, 100, 90)
        assert parse_cigar(100, '+', '10S101M10S') == (90, 100, 200, 210)
        assert parse_cigar(100, '-', '10S101M10S') == (210, 200, 100, 90)

    def test_sam_format_spec(self):
        # See 'Sequence Alignment/Map Format Specification', 2 Sep 2016.
        # Specifically, these are taken from the SAM file format specification,
        # section 1.1 ("An example").

        # +r001/1
        assert parse_cigar(7, '+', '8M2I4M1D3M') == (7, 7, 22, 22)
        # +r002
        assert parse_cigar(9, '+', '3S6M1P1I4M') == (6, 9, 18, 18)
        # +r003
        assert parse_cigar(9, '+', '5S6M') == (4, 9, 14, 14)
        # +r004
        assert parse_cigar(16, '+', '6M14N5M') == (16, 16, 40, 40)
        # -r003, hard clipping, we don't support it...
        with pytest.raises(HardClippingNotSupportedException,
                message='Dupligänger does not support hard-clipping. cigar: 6H5M, left pos: 29, strand: -'):
            parse_cigar(29, '-', '6H5M') == (16, 16, 40, 40)
        # -r001/2
        assert parse_cigar(37, '-', '9M') == (45, 45, 37, 37)

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

"""Remove adapters.  This is simply a 'Cutadapt' wrapper.

Usage:
    dupliganger remove-adapter [options] <input.fastq>
    dupliganger remove-adapter [options] <in1.fastq> <in2.fastq>


Note:
    Dupligänger supports (and autodetects) input FASTQ files that are gzipped.

Options:
    -h, --help
    -v, --verbose   Be verbose
    --compress      Compress output.
    -o OUT_DIR      Place results in directory OUT_DIR.

    -C S, --cutadapt S  Pass 'S' as arguments to Cutadapt. Don't forget to put quotes
                        around S if passing in more than one argument.
                        [default: '-n 3 -O 1 -m 30']
    -1 A, --adapter1    First (Illumina) adapter [default: AGATCGGAAGAGC]
    -2 A, --adapter2    Second (Illumina) adapter [default: AGATCGGAAGAGC]

"""


###############
### Imports ###
###############

# Python 2/3 compatibility imports
from __future__ import absolute_import, division, print_function

# NOTE: Do *not* do the following:
# from builtins import str, chr, object

# Dupligänger imports
from dupliganger.constants import *
from dupliganger.exceptions import *
from dupliganger.common import (pgopen, tmpf_start, tmpf_finish,
        filename_in_to_out_fqgz, pe_log_filename, se_log_filename,
        args_to_out_dir)

## Other imports
from docopt import docopt

# For converting ~ to full path
import os

# # For shell-like "which()"
try:
    from shutil import which
except ImportError:
    from whichcraft import which

# For external command execution
import subprocess, shlex


#################
### Constants ###
#################


#################
### Functions ###
#################

def pe_remove_adapters(in1, in2, out1, out2, short1, short2, out_log, adapter1,
        adapter2, cutadapt_args):
    """Remove adapter, paired-end version.

    Args:
        in1 (file): Read1 fastq input file.
        in2 (file): Read1 fastq input file.
        out1 (file): Read1 fastq adapter removed output file.
        out2 (file): Read2 fastq adapter removed output file.
        short1 (file): Read1 cutadapt --too-short-output file.
        short2 (file): Read2 cutadapt --too-short-paired-output file.
        out_log (str): Stdout/stderr of cutadapt execution.
        adapter1 (str): Illumina adapter1.
        adapter2 (str): Illumina adapter2.
        cutadapt_args (str): Arguments to pass to cutadapt.
    """

    adapters = "-a {} -A {}".format(adapter1, adapter2)
    ins = "{} {}".format(in1, in2)
    outs = "-o {} -p {}".format(out1, out2)
    too_short = "--too-short-output={} --too-short-paired-output={}".format(
            short1, short2)

    cmd = "cutadapt {} {} {} {} {}".format(cutadapt_args, adapters, ins, outs,
            too_short)
    print("Running:\n\t{}".format(cmd))
    cmd = shlex.split(cmd)

    with open(out_log, 'w') as f:
        p = subprocess.Popen(cmd, stdout=f, stderr=subprocess.STDOUT, bufsize=-1)
        retval = p.wait()
    if retval > 0:
        raise ExecutionException(
                """Failure to execute properly. See {}""".format(out_log))

def se_remove_adapter(in1, out1, short1, out_log, adapter1, cutadapt_args):
    """Remove adapter, single-end version.

    Args:
        in1 (file): fastq input file.
        out1 (file): fastq adapter removed output file.
        short1 (file): cutadapt --too-short-output file.
        out_log (str): Stdout/stderr of cutadapt execution.
        adapter1 (str): Illumina adapter1.
        cutadapt_args (str): Arguments to pass to cutadapt.
    """

    adapters = "-a {}".format(adapter1)
    ins = "{}".format(in1)
    outs = "-o {}".format(out1)
    too_short = "--too-short-output={}".format(short1)

    cmd = "cutadapt {} {} {} {} {}".format(cutadapt_args, adapters, ins, outs,
            too_short)
    print("Running:\n\t{}".format(cmd))
    cmd = shlex.split(cmd)

    with open(out_log, 'w') as f:
        p = subprocess.Popen(cmd, stdout=f, stderr=subprocess.STDOUT, bufsize=-1)
        retval = p.wait()
    if retval > 0:
        raise ExecutionException(
                """Failure to execute properly. See {}""".format(out_log))

def parse_args(args):
    """Parse the command line arguments."""
    adapters = []

    # Convert ~ to real path
    if args['<in2.fastq>']:
        in1 = os.path.expanduser(args['<in1.fastq>'])
        in2 = os.path.expanduser(args['<in2.fastq>'])
        input_files = [in1, in2]
        if not args['--adapter1'] or not args['--adapter2']:
            raise ArgumentException(
                    """Error:

                    --adapter1 and --adapter2 are required if running in
                    paired-end mode (i.e. when you give two FASTQ files).

                    """)
        adapters = [args['--adapter1'], args['--adapter2']]
    else:
        input_files = [os.path.expanduser(args['<input.fastq>'])]
        adapters = [args['--adapter1']]
        # Uncomment if we ever remove the default from the --apapter arguments
        # if args['--adapter2']:
        #     raise ArgumentException(
        #             """Error:

        #             --adapter2 was given (suggesting you wanted paired-end
        #             adapter removal), but only one FASTQ file was given.

        #             """)

    outdir = args_to_out_dir(args)

    cutadapt_args = args['--cutadapt'].strip("'").strip('"').strip("'")

    # compress output?
    compress = args['--compress']
    return (compress, outdir, cutadapt_args, adapters, input_files)

def run(compress, outdir, cutadapt_args, adapters, input_files):
    """Start the run.

    Args:
        compress (bool): Whether or not to compress output.
        outdir (str): Output directory for results
        cutadapt_args (str): Arguments to pass to cutadapt.
        adapters ([str]): List of Illumina adapters to cutadapt.
        input_files ([str]): Array of input fastq file(s) to be parsed.
    Returns:
        out_files ([str]): Array of output fastq file(s).

    """
    if not which('cutadapt'):
        raise PrerequisitesException(
                """Cannot find 'cutadapt'.  Install with:

                pip install cutadapt

                """)

    if len(input_files) == 2:
        in1, in2 = input_files
        out1 = filename_in_to_out_fqgz(in1, SUFFIX_REMOVE_ADAPTER, compress,
                outdir)
        out2 = filename_in_to_out_fqgz(in2, SUFFIX_REMOVE_ADAPTER, compress,
                outdir)
        out_files = [out1, out2]
        adapter1, adapter2 = adapters
        out_log = pe_log_filename(SUFFIX_REMOVE_ADAPTER, out2)
        # i.e. for those that are too short...
        short_out1 = filename_in_to_out_fqgz(out1,
                SUFFIX_REMOVE_ADAPTER_TOO_SHORT, compress, outdir)
        short_out2 = filename_in_to_out_fqgz(out2,
                SUFFIX_REMOVE_ADAPTER_TOO_SHORT, compress, outdir)

        tmp_out1, tmp_out2, tmp_short_out1, tmp_short_out2 = tmpf_start(out1,
                out2, short_out1, short_out2)

        pe_remove_adapters(in1, in2, tmp_out1, tmp_out2, tmp_short_out1,
                tmp_short_out2, out_log, adapter1, adapter2, cutadapt_args)
        tmpf_finish(tmp_out1, tmp_out2, tmp_short_out1, tmp_short_out2)

    elif len(input_files) == 1:
        in1 = input_files[0]
        out1 = filename_in_to_out_fqgz(in1, SUFFIX_REMOVE_ADAPTER, compress,
                outdir)
        out_files = [out1]
        adapter1 = adapters[0]
        out_log = se_log_filename(SUFFIX_REMOVE_ADAPTER, out1)
        # i.e. for those that are too short...
        short_out1 = filename_in_to_out_fqgz(out1,
                SUFFIX_REMOVE_ADAPTER_TOO_SHORT, compress, outdir)

        tmp_out1, tmp_short_out1 = tmpf_start(out1, short_out1)

        se_remove_adapter(in1, tmp_out1, tmp_short_out1, out_log, adapter1,
                cutadapt_args)
        tmpf_finish(tmp_out1, tmp_short_out1)

    else:
        raise ControlFlowException(
                """ERR911: Not possible to be here.""")

    return out_files


###############
### Classes ###
###############


############
### Main ###
############

def main():
    args = docopt(__doc__)
    run(*parse_args(args))

# vim: softtabstop=4:shiftwidth=4:expandtab
